// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  crystal growth group                                                  ==
// ==                                                                        ==
// ==  Stiftung caesar                                                       ==
// ==  Ludwig-Erhard-Allee 2                                                 ==
// ==  53175 Bonn                                                            ==
// ==  germany                                                               ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  http://www.caesar.de/cg/AMDiS                                         ==
// ==                                                                        ==
// ============================================================================

/** \file ILUTPreconditioner.h */

#ifndef AMDIS_ILUT_PRECONDITIONER_H
#define AMDIS_ILUT_PRECONDITIONER_H

#include <vector>
#include "Preconditioner.h"
#include "MemoryManager.h"
#include "CreatorInterface.h"
#include "DOFMatrix.h"
#include "Parameters.h"

namespace AMDiS {

  template<typename T> class DOFVector;
  template<typename T> class OEMSolver;

  // ============================================================================
  // ===== class ILUTPreconditioner ==============================================
  // ============================================================================

  /**
   * \ingroup Solver
   * 
   * \brief
   * ILUT (Incomplete LU factorization with Threshold) preconditioner.
   *
   * The preconditioner is implemented following the book "Iterative Methods for
   * Sparce Linear Systems", second edition, Yousef Saad. The preconditioner is
   * described in chapter 10.4 (algorithm 10.6).
   * We have changed the dropping rules of the original algorithm to that the
   * absolut values of w_k are checked.
   */
  class ILUTPreconditioner : public PreconditionerScal
  {
  public:
    MEMORY_MANAGED(ILUTPreconditioner);

    /** \brief
     * Creater class used in PreconditionerMap.
     */     
    class Creator : public PreconditionerScalCreator
    {
    public:
      MEMORY_MANAGED(Creator);

      /** \brief
       * Creates an ILUPreconditioner.
       */
      PreconditionerScal *create() {
	return NEW ILUTPreconditioner(this->name_, size, row);
      };
    };

    /** \brief
     * Constructor.
     */
    ILUTPreconditioner(std::string name, int size = 1, int row = 0)
      : PreconditionerScal(size, row),
	name_(name),
	tau_(0.0001),
	p_(8)
    {
      GET_PARAMETER(0, name + "->tau", "%f", &tau_);      
      GET_PARAMETER(0, name + "->p", "%d", &p_);      
    };

    /** \brief
     * Destructor.
     */
    virtual ~ILUTPreconditioner() {};

    /** \brief
     * Realisation of Preconditioner::init
     *
     */
    void init();

    /** \brief
     * Realisation of Preconditioner::exit
     */
    void exit();
    
    /** \brief
     * Realisation of Preconditioner::precon
     */
    void precon(DOFVector<double>* x);

  private:
    std::string name_;

    /** \brief
     * This matrix stores the incomplete LU factorization
     * of the coefficient matrix.
     */
    DOFMatrix ilu_;

    /** \brief
     * Here we store the diagonal elements of ilu_ matrix. This
     * allows faster computations.
     */
    std::vector<double> diag_;

    /** \brief
     * Stores the threshold factor for the preconditioning process.
     */
    double tau_;
 
    /** \brief
     * Stores the number of entries that are finally stored in each row
     * of L and U. Hence, the total number of elements in each row of ilu_ 
     * is 2p + 1 (the diagonal element is always stored).
     */
    int p_;
  };
}

#endif // AMDIS_ILUT_PRECONDITIONER_H
