// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  crystal growth group                                                  ==
// ==                                                                        ==
// ==  Stiftung caesar                                                       ==
// ==  Ludwig-Erhard-Allee 2                                                 ==
// ==  53175 Bonn                                                            ==
// ==  germany                                                               ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  http://www.caesar.de/cg/AMDiS                                         ==
// ==                                                                        ==
// ============================================================================

/** \file SMIAdapter.h */

#ifndef AMDIS_SMIADAPTER_H
#define AMDIS_SMIADAPTER_H

#include "Global.h"
#include "smi.h"
#include "MemoryManager.h"
#include "Flag.h"
#include <map>
#include <vector>

namespace AMDiS {

  class FiniteElemSpace;
  class ElInfo;
  template<typename T> class DOFVector;

#define SMI_NEIGH_QUANTITY 10001
#define SMI_OPP_V_QUANTITY 20002

  // ============================================================================
  // ===== class SMIAdapter =====================================================
  // ============================================================================

  /** \brief
   * SMI adapter for an AMDiS mesh.
   */
  class SMIAdapter
  {
  public:
    MEMORY_MANAGED(SMIAdapter);

    /** \brief
     * constructor
     */
    SMIAdapter(int              smiApplicationId,
	       int              smiMeshId,
	       FiniteElemSpace *feSpace,
	       int              elementRegion = -1,
	       int              surfaceRegion = -1,
	       bool (*elementFct)(ElInfo *elInfo) = NULL,
	       bool (*surfaceFct)(ElInfo *elInfo, int side) = NULL);

    /** \brief
     * destructor
     */
    virtual ~SMIAdapter() {};

    /** \brief
     * Add DOFVector<double> as smi quantity.
     */
    void addQuantity(int quantityId, DOFVector<double> *dofVector);

    /** \brief
     * Add DOFVector<double> as smi quantity.
     */
    void addQuantity(int quantityId, 
		     int quantityDim,
		     DOFVector<double> **dofVector);

    /** \brief
     * If called before the first call of \ref transferMeshToSMI() 
     * neighbourhood information are stored in SMI as quantities.
     */
    void addNeighbourInfo();

    /** \brief
     * Get neighbourhood information about the element. 
     * \param neighbours array of the neighbour element indices
     * \param oppVertices array of the corresponding opp vertices 
     */
    void getNeighbourInfo(int  elementIndex, 
			  int *neighbours, 
			  int *oppVertices);

    /** \brief
     * Transfer mesh data from AMDiS to SMI.
     */
    void transferMeshToSMI();

    /** \brief
     * Transfer quantity values from AMDiS to SMI.
     */
    void transferQuantitiesToSMI(int quantityID = -1);

    /** \brief
     * Transfer quantity values from SMI to AMDiS.
     */
    void getQuantitiesFromSMI(int quantityID = -1);

    void setIndexMappings(std::map<DegreeOfFreedom, DegreeOfFreedom> *newNodeIndex,
			  std::map<DegreeOfFreedom, DegreeOfFreedom> *oldNodeIndex,
			  std::map<int, int> *newElementIndex,
			  std::map<int, int> *oldElementIndex)
    {
      newNodeIndex_ = newNodeIndex;
      oldNodeIndex_ = oldNodeIndex;
      newElementIndex_ = newElementIndex;
      oldElementIndex_ = oldElementIndex;
    };

  protected:
    /** \brief
     * SMI application id
     */
    int smiApplicationId_;

    /** \brief
     * SMI mesh id
     */
    int smiMeshId_;

    /** \brief
     * FESpace containing the mesh
     */
    FiniteElemSpace *feSpace_;

    /** \brief
     * smi element type of all elements in the mesh 
     */
    int elementType_;

    /** \brief
     * mapping from quantity ids to DOFVector pointers 
     */
    std::map<int, std::vector<DOFVector<double>*> > dofVectors_;

    /** \brief
     * element region the adapter is defined on
     */
    int elementRegion_;

    /** \brief
     * surface region the adapter is defined on
     */
    int surfaceRegion_;

    /** \brief
     * store neighbourhood information?
     */
    bool addNeighbourInfo_;

    std::map<DegreeOfFreedom, DegreeOfFreedom> *newNodeIndex_;
    std::map<DegreeOfFreedom, DegreeOfFreedom> *oldNodeIndex_;
    std::map<int, int> *newElementIndex_;
    std::map<int, int> *oldElementIndex_;

    bool (*elementFct_)(ElInfo*);
    bool (*surfaceFct_)(ElInfo*, int);
  };

}

#endif
