#include "AMDiS.h"
#include "MpCCIAdapter.h"

using namespace std;
using namespace AMDiS;

// ===== function definitions // 
/** \brief
 * Dirichlet boundary function
 */
class G : public AbstractFunction<double, WorldVector<double> >
{
public:
  MEMORY_MANAGED(G);

  /** \brief
   * Implementation of AbstractFunction::operator().
   */
  double operator()(const WorldVector<double>& x) const {
    return exp(-10.0 * (x * x));
  }
};

/** \brief
 * RHS function
 */
class F : public AbstractFunction<double, WorldVector<double> >
{
public:
  MEMORY_MANAGED(F);

  F(int degree) : AbstractFunction<double, WorldVector<double> >(degree) {};

  /** \brief
   * Implementation of AbstractFunction::operator().
   */
  double operator()(const WorldVector<double>& x) const {
    int dim = x.getSize();
    double r2 = (x * x);
    double ux = exp(-10.0 * r2);
    return -(400.0 * r2 - 20.0 * dim) * ux;
  }
};

// // ===== main program // 
int main(int argc, char* argv[])
{
  FUNCNAME("main");

  // ===== check for init file =====
  TEST_EXIT(argc == 3)("usage: mpccitest initfile codeID\n");

  int codeID = atoi(argv[2]);

  // ===== init parameters =====
  Parameters::init(false, argv[1]);

  // ===== create and init the scalar problem ===== 
  ProblemScal mpccitest("mpccitest");

  mpccitest.initialize(INIT_ALL);

  // === create adapt info ===
  AdaptInfo *adaptInfo = NEW AdaptInfo("mpccitest->adapt", 1);

  // === create adapt ===
  AdaptStationary *adapt = NEW AdaptStationary("mpccitest->adapt",
					       &mpccitest,
					       adaptInfo);
  
  //mpccitest.setAdapt(adapt);

  // ===== add boundary conditions =====
  mpccitest.addDirichletBC(1, NEW G);

  // ===== create matrix operator =====
  Operator matrixOperator(Operator::MATRIX_OPERATOR, mpccitest.getFESpace());
  matrixOperator.addSecondOrderTerm(NEW Laplace_SOT);
  mpccitest.addMatrixOperator(&matrixOperator);

  // ===== create rhs operator =====
  int degree = mpccitest.getFESpace()->getBasisFcts()->getDegree();
  Operator rhsOperator(Operator::VECTOR_OPERATOR, mpccitest.getFESpace());
  rhsOperator.addZeroOrderTerm(NEW CoordsAtQP_ZOT(NEW F(degree)));
  mpccitest.addVectorOperator(&rhsOperator);

  mpccitest.getMesh()->setPreserveCoarseDOFs(true);

  
  // ===== start adaption loop =====
  if(codeID == 1)
    adapt->adapt();


  DOFVector<double> vec1(mpccitest.getFESpace(), "vec1");
  DOFVector<double> vec2(mpccitest.getFESpace(), "vec2");

  int remoteCodeID;
  int convergence;
  int syncPointID = 1;
  int quantityIDs[2] = {1, 2};
  DOFVector<double> *vectors[2] = {&vec1, &vec2};

  if(codeID == 1) {
    MpCCIAdapter mpcciAdapter(123, 1, 1, 
			      mpccitest.getFESpace(), 1, -1,
			      2, quantityIDs, vectors,
			      1, &syncPointID);
    mpcciAdapter.remesh();
    mpcciAdapter.checkConvergence(CCI_CONTINUE, &convergence);

    remoteCodeID = 2;
    vec1.set(1.0);
    vec2.set(0.0);
    mpcciAdapter.putNodes(codeID);
    mpcciAdapter.send(1, &codeID);
    mpcciAdapter.recv(1, &remoteCodeID);
    mpcciAdapter.getNodes(remoteCodeID);

    vec1.print();
    vec2.print();

    WAIT;

    vec1.set(1.0);
    vec2.set(0.0);

    mpcciAdapter.putNodes(codeID);
    mpcciAdapter.reachSyncPoint(1);
    mpcciAdapter.getNodes(remoteCodeID);

    vec1.print();
    vec2.print();
  } 
  if(codeID == 2) {
    MpCCIAdapter mpcciAdapter(124, 1, 1, 
			      mpccitest.getFESpace(), 1, -1, 
			      2, quantityIDs, vectors,
			      1, &syncPointID);
    mpcciAdapter.checkConvergence(CCI_CONTINUE, &convergence);

    remoteCodeID = 1;
    vec1.set(0.0);
    vec2.set(2.0);
    mpcciAdapter.putNodes(codeID);
    mpcciAdapter.recv(1, &remoteCodeID);
    mpcciAdapter.send(1, &codeID);
    mpcciAdapter.getNodes(remoteCodeID);

    vec1.print();
    vec2.print();

    WAIT;

    vec1.set(0.0);
    vec2.set(2.0);

    mpcciAdapter.putNodes(codeID);
    mpcciAdapter.reachSyncPoint(1);
    mpcciAdapter.getNodes(remoteCodeID);

    vec1.print();
    vec2.print();
  }

}


